import contextlib
import random
import time
import typing
from collections import deque

from pocket_option.constants import TIMESTAMP_OFFSET
from pocket_option.types import JsonFunction, JsonValue

__all__ = ("append_or_replace", "fix_timestamp", "generate_request_id", "get_json_function")

rnd = random.SystemRandom()


def get_json_function() -> JsonFunction:
    with contextlib.suppress(ImportError):
        import ujson  # type: ignore  # noqa: PLC0415

        class _UJson:
            def loads(self, value: str | bytes) -> JsonValue:
                return ujson.loads(value)

            def dumps(self, value: JsonValue, *, separators: tuple[str, str] | None = None) -> str:
                return ujson.dumps(value, ensure_ascii=False, separators=separators)

        return _UJson()

    import json  # noqa: PLC0415

    class _JsonLoads:
        def loads(self, value: str | bytes) -> JsonValue:
            return json.loads(value)

        def dumps(self, value: JsonValue, *, separators: tuple[str, str] | None = None) -> str:
            return json.dumps(value, ensure_ascii=False, separators=separators)

    return _JsonLoads()


def fix_timestamp(ts: float) -> float:
    return ts + TIMESTAMP_OFFSET


@typing.overload
def append_or_replace[T](
    array: list[T],
    item: T,
    eq_by_keys: list[str],
    get_key_method: typing.Callable[[T, str], typing.Any] = getattr,
) -> list[T]: ...
@typing.overload
def append_or_replace[T](
    array: deque[T],
    item: T,
    eq_by_keys: list[str],
    get_key_method: typing.Callable[[T, str], typing.Any] = getattr,
) -> deque[T]: ...
def append_or_replace[T](
    array: list[T] | deque[T],
    item: T,
    eq_by_keys: list[str],
    get_key_method: typing.Callable[[T, str], typing.Any] = getattr,
) -> list[T] | deque[T]:
    for i, it in enumerate(array):
        if all(get_key_method(it, key) == get_key_method(item, key) for key in eq_by_keys):
            array[i] = item
            return array
    array.append(item)
    return array


def generate_request_id() -> int:
    return int(time.time()) - TIMESTAMP_OFFSET + rnd.randint(1, 100)



import collections.abc
import typing

__all__ = ("EmitCallback", "JsonFunction", "JsonValue", "SIOEventListener", "TypedEventListener")


type JsonValue = int | float | str | "dict[str, JsonValue]" | "list[JsonValue]"


class JsonFunction(typing.Protocol):
    def dumps(self, value: JsonValue, *, separators: tuple[str, str] | None = None) -> JsonValue: ...
    def loads(self, value: str | bytes) -> JsonValue: ...


type EmitCallback[T] = (
    collections.abc.Callable[
        [str, int, T],
        None,
    ]
    | collections.abc.Callable[
        [str, int, T],
        collections.abc.Coroutine[None, None, None],
    ]
)
type SIOEventListener = (
    collections.abc.Callable[..., None] | collections.abc.Callable[..., collections.abc.Coroutine[None, None, None]]
)

type TypedEventListener[T] = (
    collections.abc.Callable[[T], None] | collections.abc.Callable[[T], collections.abc.Coroutine[None, None, None]]
)



import abc
import collections.abc
import datetime
import math
import typing
from collections import defaultdict, deque

import pydantic
import pytz

from pocket_option.generated_client import PocketOptionClient
from pocket_option.models import Asset, UpdateStreamItem
from pocket_option.utils import append_or_replace

if typing.TYPE_CHECKING:
    from pocket_option.generated_client import PocketOptionClient


class Candle(pydantic.BaseModel):
    asset: Asset
    timestamp: datetime.datetime
    timeframe: int
    open: float
    low: float
    high: float
    close: float


class CandleStorage(abc.ABC):
    def __init__(self, client: "PocketOptionClient") -> None:
        self.client = client

        self.client.on.update_stream(self._on_update_stream)

    async def _on_update_stream(self, items: list[UpdateStreamItem]) -> None:
        await self.add_item_bulk(items)

    async def add_candle(self, candle: Candle) -> None:
        await self.add_item_bulk(
            [
                UpdateStreamItem(asset=candle.asset, timestamp=candle.timestamp.timestamp(), value=candle.open),
                UpdateStreamItem(asset=candle.asset, timestamp=candle.timestamp.timestamp() + 0.01, value=candle.low),
                UpdateStreamItem(asset=candle.asset, timestamp=candle.timestamp.timestamp() + 0.02, value=candle.high),
                UpdateStreamItem(
                    asset=candle.asset,
                    timestamp=candle.timestamp.timestamp() + (candle.timeframe - 0.01),
                    value=candle.close,
                ),
            ],
        )

    @abc.abstractmethod
    async def add_item(self, item: UpdateStreamItem): ...
    @abc.abstractmethod
    async def add_item_bulk(self, items: list[UpdateStreamItem]): ...
    @abc.abstractmethod
    async def get_items(
        self,
        asset: Asset,
        *,
        start: datetime.datetime | None = None,
        end: datetime.datetime | None = None,
        count: int | None = None,
    ) -> collections.abc.Iterable[UpdateStreamItem]: ...

    async def get_candles(
        self,
        asset: Asset,
        timeframe: int = 5,
        *,
        start: datetime.datetime | None = None,
        end: datetime.datetime | None = None,
        count: int | None = None,
    ) -> collections.abc.Iterable[Candle]:
        items = await self.get_items(asset, start=start, end=end, count=count)

        buckets: dict[int, list[UpdateStreamItem]] = defaultdict(list)
        for item in items:
            ts_bucket = math.floor(item.timestamp / timeframe) * timeframe
            buckets[ts_bucket].append(item)
        candles = []

        for ts_bucket in sorted(buckets):
            group = buckets[ts_bucket]
            values = [i.value for i in group]
            candle = Candle(
                asset=asset,
                timestamp=datetime.datetime.fromtimestamp(ts_bucket, tz=pytz.UTC),
                timeframe=timeframe,
                open=values[0],
                close=values[-1],
                high=max(values),
                low=min(values),
            )
            candles.append(candle)

        return candles


class MemoryCandleStorage(CandleStorage):
    def __init__(self, client: "PocketOptionClient") -> None:
        super().__init__(client)
        self._max_len = 10_000
        self._storage: dict[Asset, deque[UpdateStreamItem]] = defaultdict(lambda: deque([], 10_000))

    def set_max_len(self, _max_len: int):
        self._storage = defaultdict(lambda: deque([], _max_len))

    async def add_item(self, item: UpdateStreamItem):
        self._storage[item.asset] = append_or_replace(self._storage[item.asset], item, ["asset", "timestamp"])

    async def add_item_bulk(self, items: list[UpdateStreamItem]):
        for it in items:
            await self.add_item(it)

    async def get_items(
        self,
        asset: Asset,
        *,
        start: datetime.datetime | None = None,
        end: datetime.datetime | None = None,
        count: int | None = None,
    ) -> collections.abc.Iterable[UpdateStreamItem]:
        items = self._storage.get(asset, [])
        if start:
            items = [i for i in items if i.timestamp >= start.timestamp()]
        if end:
            items = [i for i in items if i.timestamp <= end.timestamp()]
        items = list(items)
        items.sort(key=lambda i: i.timestamp)
        if count is not None:
            items = items[-count:]
        return items



import typing

from pocket_option import models
from pocket_option.client import BasePocketOptionClient

if typing.TYPE_CHECKING:
    from pocket_option.types import TypedEventListener

__all__ = ("PocketOptionClient",)


class PocketOptionClientEmit:
    def __init__(self, client: BasePocketOptionClient) -> None:
        self.client = client

    async def ps(self) -> None:
        """No description"""
        await self.client.send("ps")

    async def indicator_load(self) -> None:
        """No description"""
        await self.client.send("indicator/load")

    async def favorite_load(self) -> None:
        """No description"""
        await self.client.send("favorite/load")

    async def price_alert_load(self) -> None:
        """No description"""
        await self.client.send("price-alert/load")

    async def auth(self, data: models.AuthorizationData) -> None:
        """No description


        :type data: models.AuthorizationData
        """
        await self.client.send("auth", data)

    async def subscribe_symbol(self, asset: models.Asset) -> None:
        """Subscribes to live updates for a specific asset symbol.

        :param asset: The asset symbol to subscribe to.
        :type asset: models.Asset
        """
        await self.client.send("subscribeSymbol", asset)

    async def subscribe_for(self, asset: models.Asset) -> None:
        """Subscribes to specific data streams for the given asset.

        :param asset: The asset symbol to subscribe for.
        :type asset: models.Asset
        """
        await self.client.send("subfor", asset)

    async def unsubscribe_for(self, asset: models.Asset) -> None:
        """Unsubscribes from data streams for the given asset.

        :param asset: The asset symbol to unsubscribe from.
        :type asset: models.Asset
        """
        await self.client.send("unsubfor", asset)

    async def change_symbol(self, data: models.ChangeSymbolRequest) -> None:
        """No description


        :type data: models.ChangeSymbolRequest
        """
        await self.client.send("changeSymbol", data)

    async def open_order(self, data: models.OpenOrderRequest) -> None:
        """Sends a request to open a new trading order.

        :param data: Order request payload.
        :type data: models.OpenOrderRequest
        """
        await self.client.send("openOrder", data)

    async def copy_signal_order(self, data: models.CopySignalRequest) -> None:
        """Executes an order based on a copied trading signal.

        :param data: Copy trading signal request payload.
        :type data: models.CopySignalRequest
        """
        await self.client.send("copySignalOrder", data)


class PocketOptionClientOn:
    def __init__(self, client: BasePocketOptionClient) -> None:
        self.client = client

    @typing.overload
    def success_update_balance(
        self,
        handler: None = None,
    ) -> "typing.Callable[[TypedEventListener[models.SuccessUpdateBalance]], None]": ...

    @typing.overload
    def success_update_balance(
        self,
        handler: "TypedEventListener[models.SuccessUpdateBalance]",
    ) -> None: ...

    def success_update_balance(
        self,
        handler: "TypedEventListener[models.SuccessUpdateBalance] | None" = None,
    ) -> "None | typing.Callable[[TypedEventListener[models.SuccessUpdateBalance]], None]":
        """Triggered when the user's account balance is updated.

        :param handler: Callback
        :type handler: TypedEventListener[models.SuccessUpdateBalance] | None
        """
        return self.client.add_on("successupdateBalance", handler=handler, model=models.SuccessUpdateBalance)

    @typing.overload
    def update_history_new_fast(
        self,
        handler: None = None,
    ) -> "typing.Callable[[TypedEventListener[models.UpdateHistoryFastEvent]], None]": ...

    @typing.overload
    def update_history_new_fast(
        self,
        handler: "TypedEventListener[models.UpdateHistoryFastEvent]",
    ) -> None: ...

    def update_history_new_fast(
        self,
        handler: "TypedEventListener[models.UpdateHistoryFastEvent] | None" = None,
    ) -> "None | typing.Callable[[TypedEventListener[models.UpdateHistoryFastEvent]], None]":
        """Triggered when a new fast trading history record is received.

        :param handler: Callback
        :type handler: TypedEventListener[models.UpdateHistoryFastEvent] | None
        """
        return self.client.add_on("updateHistoryNewFast", handler=handler, model=models.UpdateHistoryFastEvent)

    @typing.overload
    def update_stream(
        self,
        handler: None = None,
    ) -> "typing.Callable[[TypedEventListener[list[models.UpdateStreamItem]]], None]": ...

    @typing.overload
    def update_stream(
        self,
        handler: "TypedEventListener[list[models.UpdateStreamItem]]",
    ) -> None: ...

    def update_stream(
        self,
        handler: "TypedEventListener[list[models.UpdateStreamItem]] | None" = None,
    ) -> "None | typing.Callable[[TypedEventListener[list[models.UpdateStreamItem]]], None]":
        """Triggered when a live data stream update is received.

        :param handler: Callback
        :type handler: TypedEventListener[list[models.UpdateStreamItem]] | None
        """
        return self.client.add_on("updateStream", handler=handler, model=models.UpdateStreamTypeAdapter)

    @typing.overload
    def update_opened_deals(
        self,
        handler: None = None,
    ) -> "typing.Callable[[TypedEventListener[list[models.Deal]]], None]": ...

    @typing.overload
    def update_opened_deals(
        self,
        handler: "TypedEventListener[list[models.Deal]]",
    ) -> None: ...

    def update_opened_deals(
        self,
        handler: "TypedEventListener[list[models.Deal]] | None" = None,
    ) -> "None | typing.Callable[[TypedEventListener[list[models.Deal]]], None]":
        """Triggered when the list of currently opened deals is updated.

        :param handler: Callback
        :type handler: TypedEventListener[list[models.Deal]] | None
        """
        return self.client.add_on("updateOpenedDeals", handler=handler, model=models.DealListTypeAdapter)

    @typing.overload
    def success_open_order(
        self,
        handler: None = None,
    ) -> "typing.Callable[[TypedEventListener[models.Deal]], None]": ...

    @typing.overload
    def success_open_order(
        self,
        handler: "TypedEventListener[models.Deal]",
    ) -> None: ...

    def success_open_order(
        self,
        handler: "TypedEventListener[models.Deal] | None" = None,
    ) -> "None | typing.Callable[[TypedEventListener[models.Deal]], None]":
        """No description

        :param handler: Callback
        :type handler: TypedEventListener[models.Deal] | None
        """
        return self.client.add_on("successopenOrder", handler=handler, model=models.Deal)

    @typing.overload
    def update_closed_deals(
        self,
        handler: None = None,
    ) -> "typing.Callable[[TypedEventListener[list[models.Deal]]], None]": ...

    @typing.overload
    def update_closed_deals(
        self,
        handler: "TypedEventListener[list[models.Deal]]",
    ) -> None: ...

    def update_closed_deals(
        self,
        handler: "TypedEventListener[list[models.Deal]] | None" = None,
    ) -> "None | typing.Callable[[TypedEventListener[list[models.Deal]]], None]":
        """Triggered when the list of recently closed deals is updated.

        :param handler: Callback
        :type handler: TypedEventListener[list[models.Deal]] | None
        """
        return self.client.add_on("updateClosedDeals", handler=handler, model=models.DealListTypeAdapter)

    @typing.overload
    def success_close_order(
        self,
        handler: None = None,
    ) -> "typing.Callable[[TypedEventListener[models.SuccessCloseOrder]], None]": ...

    @typing.overload
    def success_close_order(
        self,
        handler: "TypedEventListener[models.SuccessCloseOrder]",
    ) -> None: ...

    def success_close_order(
        self,
        handler: "TypedEventListener[models.SuccessCloseOrder] | None" = None,
    ) -> "None | typing.Callable[[TypedEventListener[models.SuccessCloseOrder]], None]":
        """No description

        :param handler: Callback
        :type handler: TypedEventListener[models.SuccessCloseOrder] | None
        """
        return self.client.add_on("successcloseOrder", handler=handler, model=models.SuccessCloseOrder)


class PocketOptionClient(BasePocketOptionClient):
    @property
    def on(self) -> PocketOptionClientOn:
        return PocketOptionClientOn(self)

    @property
    def emit(self) -> PocketOptionClientEmit:
        return PocketOptionClientEmit(self)



import contextlib
import typing

from pocket_option.middleware import Middleware
from pocket_option.utils import fix_timestamp, get_json_function

if typing.TYPE_CHECKING:
    from pocket_option.types import JsonValue

__all__ = (
    "FixTypesOnMiddleware",
    "MakeJsonOnMiddleware",
)


class MakeJsonOnMiddleware(Middleware):
    def __init__(self) -> None:
        self.json = get_json_function()

    async def on(self, event: str, data: "str | bytes | JsonValue | None") -> "JsonValue | None":  # noqa: ARG002
        if isinstance(data, str | bytes):
            with contextlib.suppress(Exception):
                return self.json.loads(data)
        return typing.cast("JsonValue", data)


class FixTypesOnMiddleware(Middleware):
    async def on(self, event: str, data: "JsonValue | None") -> "JsonValue | None":  # type: ignore
        if data is None:
            return None
        if event == "updateStream":
            return [
                {
                    "asset": it[0],
                    "timestamp": fix_timestamp(it[1]),
                    "value": it[2],
                }
                for it in typing.cast("list[tuple[str, float, float]]", data)
            ]
        return data



import typing

if typing.TYPE_CHECKING:
    from pocket_option.types import EmitCallback, JsonValue

__all__ = ("Middleware",)


class Middleware:
    async def emit(
        self,
        event: str,
        data: "JsonValue | None" = None,
        callback: "EmitCallback[JsonValue] | None" = None,
    ) -> "tuple[str, JsonValue | None, EmitCallback[JsonValue] | None]":
        return event, data, callback

    async def on(self, event: str, data: "str | bytes | JsonValue | None") -> "JsonValue | None":  # noqa: ARG002
        return data  # type: ignore



import abc
import asyncio
import collections.abc
import contextlib
import datetime
import itertools
import typing
import uuid

from pocket_option.generated_client import PocketOptionClient
from pocket_option.models import Asset, Deal, IsDemo, OpenOrderRequest, OrderAction, SuccessCloseOrder
from pocket_option.utils import append_or_replace, generate_request_id

if typing.TYPE_CHECKING:
    from pocket_option.generated_client import PocketOptionClient


class DealsStorage:
    def __init__(self, client: "PocketOptionClient") -> None:
        self.client = client

        self._open_deal_events: dict[int, asyncio.Event] = {}
        self._close_deal_events: dict[uuid.UUID, asyncio.Event] = {}

        self.client.on.success_open_order(self._on_success_deal)
        self.client.on.success_close_order(self._on_success_close_deal)
        self.client.on.update_opened_deals(self.add_or_update_deal_bulk)
        self.client.on.update_closed_deals(self.add_or_update_deal_bulk)

    async def open_deal(
        self,
        asset: Asset,
        amount: int,
        action: OrderAction,
        time: int,
        is_demo: IsDemo = 1,
        request_id: int | None = None,
        option_type: int = 100,
    ) -> Deal:
        request_id = request_id or generate_request_id()
        self._open_deal_events[request_id] = asyncio.Event()
        await self.client.emit.open_order(
            OpenOrderRequest(
                asset=asset,
                amount=amount,
                action=action,
                time=time,
                is_demo=is_demo,
                request_id=request_id,
                option_type=option_type,
            ),
        )

        try:
            await asyncio.wait_for(self._open_deal_events[request_id].wait(), 30)
        except TimeoutError as err:
            raise TimeoutError(f"Timeout waiting for deal {request_id}") from err

        if deal := await self.get_deal(request_id=request_id):
            return deal
        raise RuntimeError(f"Failed to find deal {request_id}")

    @typing.overload
    async def check_deal_result(
        self,
        wait_time: int = ...,
        *,
        deal_id: uuid.UUID = ...,
        request_id: None = ...,
        deal: None = ...,
    ) -> Deal: ...
    @typing.overload
    async def check_deal_result(
        self,
        wait_time: int = ...,
        *,
        deal_id: None = ...,
        request_id: int = ...,
        deal: None = ...,
    ) -> Deal: ...
    @typing.overload
    async def check_deal_result(
        self,
        wait_time: int = ...,
        *,
        deal_id: None = ...,
        request_id: None = ...,
        deal: Deal = ...,
    ) -> Deal: ...
    async def check_deal_result(
        self,
        wait_time: int = 600,
        *,
        deal_id: uuid.UUID | None = None,
        request_id: int | None = None,
        deal: Deal | None = None,
    ) -> Deal:
        if not deal and (deal_id or request_id):
            deal = await self.get_deal(deal_id=deal_id, request_id=request_id)  # type: ignore
        if not deal:
            raise RuntimeError("Failed to find deal")
        self._close_deal_events[deal.id] = asyncio.Event()
        try:
            await asyncio.wait_for(self._close_deal_events[deal.id].wait(), wait_time)
        except TimeoutError as err:
            raise TimeoutError(f"Timeout waiting for deal {deal.id}") from err

        if deal := await self.get_deal(deal_id=deal.id):
            return deal
        raise RuntimeError("Failed to find deal")

    async def _on_success_deal(self, deal: Deal):
        await self.add_or_update_deal(deal)
        if deal.request_id and self._open_deal_events.get(deal.request_id):
            self._open_deal_events[deal.request_id].set()
            del self._open_deal_events[deal.request_id]

    async def _on_success_close_deal(self, close_deal: SuccessCloseOrder) -> None:
        await self.add_or_update_deal_bulk(close_deal.deals)
        for deal in close_deal.deals:
            if event := self._close_deal_events.pop(deal.id):
                event.set()

    @abc.abstractmethod
    async def add_or_update_deal(self, deal: Deal) -> None: ...
    @abc.abstractmethod
    async def add_or_update_deal_bulk(self, deals: list[Deal]) -> None: ...

    @typing.overload
    async def get_deal(
        self,
        *,
        deal_id: uuid.UUID = ...,
        request_id: None = ...,
    ) -> Deal | None: ...
    @typing.overload
    async def get_deal(
        self,
        *,
        deal_id: uuid.UUID | None = ...,
        request_id: int = ...,
    ) -> Deal | None: ...
    @abc.abstractmethod
    async def get_deal(
        self,
        *,
        deal_id: uuid.UUID | None = None,
        request_id: int | None = None,
    ) -> Deal | None: ...

    @abc.abstractmethod
    async def get_deals(
        self,
        *,
        asset: Asset | None = None,
        open_time__gt: datetime.datetime | None = None,
        open_time__gte: datetime.datetime | None = None,
        open_time__lt: datetime.datetime | None = None,
        open_time__lte: datetime.datetime | None = None,
        close_time__gt: datetime.datetime | None = None,
        close_time__gte: datetime.datetime | None = None,
        close_time__lt: datetime.datetime | None = None,
        close_time__lte: datetime.datetime | None = None,
        open_price__gt: float | None = None,
        open_price__gte: float | None = None,
        open_price__lt: float | None = None,
        open_price__lte: float | None = None,
        close_price__gt: float | None = None,
        close_price__gte: float | None = None,
        close_price__lt: float | None = None,
        close_price__lte: float | None = None,
        count: int | None = None,
    ) -> collections.abc.Iterable[Deal]: ...


class MemoryDealsStorage(DealsStorage):
    def __init__(self, client: "PocketOptionClient") -> None:
        super().__init__(client)
        self._deals: list[Deal] = []

    async def add_or_update_deal(self, deal: Deal) -> None:
        self._deals = append_or_replace(self._deals, deal, eq_by_keys=["id"])

    async def add_or_update_deal_bulk(self, deals: list[Deal]) -> None:
        for deal in deals:
            await self.add_or_update_deal(deal)

    async def get_deal(self, *, deal_id: uuid.UUID | None = None, request_id: int | None = None) -> Deal | None:
        if deal_id:
            with contextlib.suppress(StopIteration):
                return next(deal for deal in self._deals if deal.id == deal_id)

        if request_id:
            with contextlib.suppress(StopIteration):
                return next(deal for deal in self._deals if deal.request_id == request_id)
        return None

    async def get_deals(
        self,
        *,
        asset: Asset | None = None,
        open_time__gt: datetime.datetime | None = None,
        open_time__gte: datetime.datetime | None = None,
        open_time__lt: datetime.datetime | None = None,
        open_time__lte: datetime.datetime | None = None,
        close_time__gt: datetime.datetime | None = None,
        close_time__gte: datetime.datetime | None = None,
        close_time__lt: datetime.datetime | None = None,
        close_time__lte: datetime.datetime | None = None,
        open_price__gt: float | None = None,
        open_price__gte: float | None = None,
        open_price__lt: float | None = None,
        open_price__lte: float | None = None,
        close_price__gt: float | None = None,
        close_price__gte: float | None = None,
        close_price__lt: float | None = None,
        close_price__lte: float | None = None,
        count: int | None = None,
    ) -> collections.abc.Iterable[Deal]:
        def _convert_dt(dt: datetime.datetime) -> float:
            return dt.timestamp()

        data = self._deals.copy()

        if asset:
            data = filter(lambda it: it.asset == asset, data)

        if open_time__gt:
            data = filter(lambda it: _convert_dt(it.open_time) > _convert_dt(open_time__gt), data)
        if open_time__gte:
            data = filter(lambda it: _convert_dt(it.open_time) >= _convert_dt(open_time__gte), data)
        if open_time__lt:
            data = filter(lambda it: _convert_dt(it.open_time) < _convert_dt(open_time__lt), data)
        if open_time__lte:
            data = filter(lambda it: _convert_dt(it.open_time) <= _convert_dt(open_time__lte), data)
        if close_time__gt:
            data = filter(lambda it: _convert_dt(it.close_time) > _convert_dt(close_time__gt), data)
        if close_time__gte:
            data = filter(lambda it: _convert_dt(it.close_time) >= _convert_dt(close_time__gte), data)
        if close_time__lt:
            data = filter(lambda it: _convert_dt(it.close_time) < _convert_dt(close_time__lt), data)
        if close_time__lte:
            data = filter(lambda it: _convert_dt(it.close_time) <= _convert_dt(close_time__lte), data)

        if open_price__gt:
            data = filter(lambda it: it.open_price > open_price__gt, data)
        if open_price__gte:
            data = filter(lambda it: it.open_price >= open_price__gte, data)
        if open_price__lt:
            data = filter(lambda it: it.open_price < open_price__lt, data)
        if open_price__lte:
            data = filter(lambda it: it.open_price <= open_price__lte, data)
        if close_price__gt:
            data = filter(lambda it: it.close_price and it.close_price > close_price__gt, data)
        if close_price__gte:
            data = filter(lambda it: it.close_price and it.close_price >= close_price__gte, data)
        if close_price__lt:
            data = filter(lambda it: it.close_price and it.close_price < close_price__lt, data)
        if close_price__lte:
            data = filter(lambda it: it.close_price and it.close_price <= close_price__lte, data)

        data = sorted(data, key=lambda it: it.open_time)
        if count:
            data = itertools.islice(data, -count)
        return data



__all__ = (
    "API_LIMITS_MAX_CONCURRENT_ORDERS",
    "API_LIMITS_MAX_DURATION",
    "API_LIMITS_MAX_ORDER_AMOUNT",
    "API_LIMITS_MIN_DURATION",
    "API_LIMITS_MIN_ORDER_AMOUNT",
    "API_LIMITS_RATE_LIMIT",
    "DEFAULT_ORIGIN",
    "DEFAULT_USER_AGENT",
    "MAX_INT_32",
    "TIMESTAMP_OFFSET",
    "Regions",
)


API_LIMITS_MIN_ORDER_AMOUNT = 1
API_LIMITS_MAX_ORDER_AMOUNT = 50_000
API_LIMITS_MIN_DURATION = 5
API_LIMITS_MAX_DURATION = 43_200
API_LIMITS_MAX_CONCURRENT_ORDERS = 10
API_LIMITS_RATE_LIMIT = 100
MAX_INT_32 = 2_147_483_647
TIMESTAMP_OFFSET = -7200
DEFAULT_ORIGIN = "https://m.pocketoption.com"
DEFAULT_USER_AGENT = "Mozilla/5.0 (Windows NT 10.0; Win64; x64; rv:143.0) Gecko/20100101 Firefox/143.0"


class Regions:
    UNITED_STATES_NORTH = "wss://api-us-north.po.market"
    UNITED_STATES_SOUTH = "wss://api-us-south.po.market"
    EUROPA = "wss://api-eu.po.market"
    ASIA = "wss://api-asia.po.market"

    UNITED_STATES_2 = "wss://api-us2.po.market"
    UNITED_STATES_3 = "wss://api-us3.po.market"
    UNITED_STATES_4 = "wss://api-us4.po.market"

    FRANCE_1 = "wss://api-fr.po.market"
    FRANCE_2 = "wss://api-fr2.po.market"
    RUSSIA = "wss://api-msk.po.market"
    INDIA = "wss://api-in.po.market"
    FINLAND = "wss://api-fin.po.market"

    SEYCHELLES = "wss://api-sc.po.market"
    HONGKONG = "wss://api-hk.po.market"

    SERVER_1 = "wss://api-spb.po.market"
    SERVER_2 = "wss://api-l.po.market"
    SERVER_3 = "wss://api-c.po.market"

    DEMO = "wss://demo-api-eu.po.market"
    DEMO_2 = "wss://try-demo-eu.po.market"



import asyncio
import collections.abc
import logging
import typing
from inspect import isclass

import aiohttp
import pydantic
import socketio

from pocket_option.constants import DEFAULT_ORIGIN, DEFAULT_USER_AGENT
from pocket_option.middleware import Middleware
from pocket_option.middlewares import FixTypesOnMiddleware, MakeJsonOnMiddleware
from pocket_option.utils import get_json_function

if typing.TYPE_CHECKING:
    from pocket_option import models
    from pocket_option.types import EmitCallback, JsonFunction, JsonValue, SIOEventListener

__all__ = ("BasePocketOptionClient",)

logger = logging.getLogger()


class BasePocketOptionClient:
    def __init__(
        self,
        on_middlewares: "list[Middleware] | None" = None,
        *,
        reconnection: bool = True,
        reconnection_attempts: int = 0,
        reconnection_delay: float = 1.0,
        reconnection_delay_max: float = 5.0,
        randomization_factor: float = 0.5,
        logger: bool = False,
        engineio_logger: bool = False,
        json: "JsonFunction | None" = None,
        handle_sigint: bool = True,
        request_timeout: float = 5,
        http_session: aiohttp.ClientSession | None = None,
        ssl_verify: bool = True,
        websocket_extra_options: dict | None = None,
        timestamp_requests: bool = False,
    ) -> None:
        """Initializes the Socket.IO client wrapper with middleware and connection options.

        :param on_middlewares:
            A list of middlewares executed on incoming Socket.IO events.
            Each middleware can modify or intercept event data before the callback is called.
            By default, the following middlewares are applied:
            - `MakeJsonOnMiddleware()` — parses JSON data into Python objects.
            - `FixTypesOnMiddleware()` — normalizes data types for consistency.

        :param emit_middlewares:
            A list of middlewares executed before emitting events to the server.
            Each middleware can modify the outgoing payload.
            Defaults to an empty list.

        :param reconnection:
            Enables or disables automatic reconnection when the connection is lost.
            Defaults to `True`.

        :param reconnection_attempts:
            The maximum number of reconnection attempts.
            A value of `0` means unlimited retries.

        :param reconnection_delay:
            Initial delay (in seconds) before attempting to reconnect.
            Defaults to `1.0`.

        :param reconnection_delay_max:
            Maximum delay (in seconds) between reconnection attempts.
            Defaults to `5.0`.

        :param randomization_factor:
            A factor between 0 and 1 used to randomize the reconnection delay to prevent
            simultaneous reconnects.
            Defaults to `0.5`.

        :param logger:
            Enables logging for the Socket.IO client.
            Defaults to `False`.

        :param engineio_logger:
            Enables logging for the underlying Engine.IO layer.
            Defaults to `False`.

        :param json:
            Custom JSON serialization/deserialization functions.
            If not provided, a default implementation will be used.

        :param handle_sigint:
            If `True`, the client will handle SIGINT (Ctrl+C) for clean shutdown.
            Defaults to `True`.

        :param request_timeout:
            The timeout (in seconds) for HTTP requests during the Socket.IO handshake.
            Defaults to `5`.

        :param http_session:
            Optional existing `aiohttp.ClientSession` instance to reuse for HTTP requests.
            If `None`, a new session is created internally.

        :param ssl_verify:
            Whether to verify SSL certificates for secure connections.
            Defaults to `True`.

        :param websocket_extra_options:
            Optional dictionary of additional parameters for the WebSocket connection.
            Useful for setting headers or specific connection parameters.

        :param timestamp_requests:
            Whether to append a timestamp to each request for caching avoidance.
            Defaults to `False`.
        """
        self.middlewares = on_middlewares or [MakeJsonOnMiddleware(), FixTypesOnMiddleware()]
        self.json = json or get_json_function()
        self.sio = socketio.AsyncClient(
            reconnection=reconnection,
            reconnection_attempts=reconnection_attempts,
            reconnection_delay=reconnection_delay,  # pyright: ignore[reportArgumentType]
            reconnection_delay_max=reconnection_delay_max,  # pyright: ignore[reportArgumentType]
            randomization_factor=randomization_factor,
            logger=logger,
            serializer="default",
            json=self.json,
            handle_sigint=handle_sigint,
            request_timeout=request_timeout,
            http_session=http_session,
            ssl_verify=ssl_verify,
            websocket_extra_options=websocket_extra_options,
            timestamp_requests=timestamp_requests,
            engineio_logger=engineio_logger,
        )

    def get_auth_from_packet(self, packet: str) -> "models.AuthorizationData":
        packet = packet.removeprefix("42")
        json_packet = self.json.loads(packet)
        return typing.cast("models.AuthorizationData", json_packet)

    def add_middleware(self, middleware: Middleware) -> None:
        self.middlewares.append(middleware)

    async def _get_real_value[T](
        self,
        value: T
        | None
        | collections.abc.Callable[[], T]
        | collections.abc.Callable[[], collections.abc.Coroutine[None, None, T]],
    ) -> T | None:
        if callable(value):
            result = value()
            if asyncio.iscoroutine(result):
                return await result
            return result  # type: ignore
        return value

    async def wait(self):
        return await self.sio.wait()

    async def disconnect(self) -> None:
        return await self.sio.disconnect()

    async def shutdown(self) -> None:
        return await self.sio.shutdown()

    async def sleep(self, seconds: float = 0) -> None:
        return await self.sio.sleep(seconds=seconds)  # type: ignore

    async def connect(
        self,
        url: str,
        headers: dict[str, str] | collections.abc.Callable[[], dict[str, str]] | None = None,
        auth: "models.AuthorizationData | None" = None,
        wait: bool = True,
        wait_timeout: float = 1,
        retry: bool = False,
    ):
        headers = await self._get_real_value(headers) or {}
        headers.setdefault("Origin", DEFAULT_ORIGIN)
        headers.setdefault("User-Agent", DEFAULT_USER_AGENT)
        return await self.sio.connect(
            url,
            headers=headers,
            auth=auth,
            transports=["websocket"],
            namespaces=["/"],
            socketio_path="socket.io",
            wait=wait,
            wait_timeout=wait_timeout,  # type: ignore
            retry=retry,
        )

    @typing.overload
    def add_on(
        self,
        event: str,
        handler: None = ...,
        *,
        model: type[pydantic.BaseModel] | pydantic.TypeAdapter | None = ...,
    ) -> "typing.Callable[[SIOEventListener], None]": ...
    @typing.overload
    def add_on(
        self,
        event: str,
        handler: "SIOEventListener",
        *,
        model: type[pydantic.BaseModel] | pydantic.TypeAdapter | None = ...,
    ) -> None: ...

    def add_on(
        self,
        event: str,
        handler: "SIOEventListener | None" = None,
        *,
        model: type[pydantic.BaseModel] | pydantic.TypeAdapter | None = None,
    ) -> "None | typing.Callable[[SIOEventListener], None]":
        def _get_data(d: "JsonValue | None"):
            if d and isinstance(d, dict) and model and isclass(model) and issubclass(model, pydantic.BaseModel):
                return model.model_validate(d)
            if d and isinstance(model, pydantic.TypeAdapter):
                return model.validate_python(d)
            return d

        if handler is None:

            def set_handler(_handler: "SIOEventListener"):
                async def wrapper(data: "JsonValue | None"):
                    for middleware in self.middlewares:
                        data = await middleware.on(event, data)
                    new_data = _get_data(data)
                    logger.debug("New event '%s' with data %r", event, new_data)
                    result = _handler(new_data)
                    if asyncio.iscoroutine(result):
                        await result

                self.sio.on(event, handler=wrapper)

            return set_handler

        async def _handler(data: "JsonValue | None"):
            for middleware in self.middlewares:
                data = await middleware.on(event, data)
            new_data = _get_data(data)
            logger.debug("New event '%s' with data %r", event, new_data)
            result = handler(new_data)
            if asyncio.iscoroutine(result):
                await result

        return self.sio.on(event, handler=_handler)

    async def send(
        self,
        event: str,
        data: "JsonValue | pydantic.BaseModel | None" = None,
        callback: "EmitCallback[JsonValue] | None" = None,
    ) -> None:
        if isinstance(data, pydantic.BaseModel):
            data = data.model_dump(mode="json", by_alias=True)
        if isinstance(data, list):
            data = [
                it.model_dump(mode="json", by_alias=True) if isinstance(it, pydantic.BaseModel) else it for it in data
            ]
        for middleware in self.middlewares:
            event, data, callback = await middleware.emit(event, data=data, callback=callback)
        logger.debug(
            "Emitting event '%s' with data %r",
            event,
            data if event != "auth" else {**typing.cast("dict[str, JsonValue]", data), "session": "***"},
        )
        return await self.sio.emit(event=event, data=data, callback=callback)



import datetime
import enum
import typing
import uuid

import pydantic

__all__ = (
    "Asset",
    "AuthorizationData",
    "ChangeSymbolRequest",
    "Command",
    "CopySignalRequest",
    "Deal",
    "IsDemo",
    "OpenOrderRequest",
    "OpenPendingOrderRequest",
    "OpenPendingOrderRequestOpenType",
    "OrderAction",
    "SuccessUpdateBalance",
    "UpdateHistoryFastEvent",
    "UpdateStreamItem",
    "UpdateStreamTypeAdapter",
)

type IsDemo = typing.Literal[0, 1]


class Asset(str, enum.Enum):
    AUDCAD = "AUDCAD"
    EURUSD = "EURUSD"
    GBPUSD = "GBPUSD"
    USDJPY = "USDJPY"
    USDCHF = "USDCHF"
    USDCAD = "USDCAD"
    AUDUSD = "AUDUSD"
    NZDUSD = "NZDUSD"
    XAUUSD = "XAUUSD"
    XAGUSD = "XAGUSD"
    UKBrent = "UKBrent"
    USCrude = "USCrude"
    XNGUSD = "XNGUSD"
    XPTUSD = "XPTUSD"
    XPDUSD = "XPDUSD"
    BTCUSD = "BTCUSD"
    ETHUSD = "ETHUSD"
    DASH_USD = "DASH_USD"
    BTCGBP = "BTCGBP"
    BTCJPY = "BTCJPY"
    BCHEUR = "BCHEUR"
    BCHGBP = "BCHGBP"
    BCHJPY = "BCHJPY"
    DOTUSD = "DOTUSD"
    LNKUSD = "LNKUSD"
    SP500 = "SP500"
    NASUSD = "NASUSD"
    DJI30 = "DJI30"
    JPN225 = "JPN225"
    D30EUR = "D30EUR"
    E50EUR = "E50EUR"
    F40EUR = "F40EUR"
    E35EUR = "E35EUR"
    A_100GBP = "100GBP"
    AUS200 = "AUS200"
    CAC40 = "CAC40"
    AEX25 = "AEX25"
    SMI20 = "SMI20"
    H33HKD = "H33HKD"
    AAPL = "#AAPL"
    MSFT = "#MSFT"
    TSLA = "#TSLA"
    FB = "#FB"
    NFLX = "#NFLX"
    INTC = "#INTC"
    BA = "#BA"
    JPM = "#JPM"
    JNJ = "#JNJ"
    PFE = "#PFE"
    XOM = "#XOM"
    AXP = "#AXP"
    MCD = "#MCD"
    CSCO = "#CSCO"
    CITI = "#CITI"
    TWITTER = "#TWITTER"
    BABA = "#BABA"

    XAUUSD_otc = "XAUUSD_otc"
    XAGUSD_otc = "XAGUSD_otc"
    UKBrent_otc = "UKBrent_otc"
    USCrude_otc = "USCrude_otc"
    XNGUSD_otc = "XNGUSD_otc"
    XPTUSD_otc = "XPTUSD_otc"
    XPDUSD_otc = "XPDUSD_otc"
    SP500_otc = "SP500_otc"
    NASUSD_otc = "NASUSD_otc"
    DJI30_otc = "DJI30_otc"
    JPN225_otc = "JPN225_otc"
    D30EUR_otc = "D30EUR_otc"
    E50EUR_otc = "E50EUR_otc"
    F40EUR_otc = "F40EUR_otc"
    E35EUR_otc = "E35EUR_otc"
    A_100GBP_otc = "100GBP_otc"
    AUS200_otc = "AUS200_otc"
    EURRUB_otc = "EURRUB_otc"
    USDRUB_otc = "USDRUB_otc"
    EURHUF_otc = "EURHUF_otc"
    CHFNOK_otc = "CHFNOK_otc"
    Microsoft_otc = "Microsoft_otc"
    Facebook_OTC = "Facebook_OTC"
    Tesla_otc = "Tesla_otc"
    Boeing_OTC = "Boeing_OTC"
    American_Express_otc = "American_Express_otc"
    EURUSD_otc = "EURUSD_otc"
    GBPUSD_otc = "GBPUSD_otc"
    USDJPY_otc = "USDJPY_otc"
    USDCHF_otc = "USDCHF_otc"
    USDCAD_otc = "USDCAD_otc"
    AUDUSD_otc = "AUDUSD_otc"
    AUDNZD_otc = "AUDNZD_otc"
    AUDCAD_otc = "AUDCAD_otc"
    AUDCHF_otc = "AUDCHF_otc"
    AUDJPY_otc = "AUDJPY_otc"
    CADCHF_otc = "CADCHF_otc"
    CADJPY_otc = "CADJPY_otc"
    CHFJPY_otc = "CHFJPY_otc"
    EURCHF_otc = "EURCHF_otc"
    EURGBP_otc = "EURGBP_otc"
    EURJPY_otc = "EURJPY_otc"
    EURNZD_otc = "EURNZD_otc"
    GBPAUD_otc = "GBPAUD_otc"
    GBPJPY_otc = "GBPJPY_otc"
    NZDJPY_otc = "NZDJPY_otc"
    NZDUSD_otc = "NZDUSD_otc"
    AAPL_otc = "#AAPL_otc"
    MSFT_otc = "#MSFT_otc"
    TSLA_otc = "#TSLA_otc"
    FB_otc = "#FB_otc"
    AMZN_otc = "#AMZN_otc"
    NFLX_otc = "#NFLX_otc"
    INTC_otc = "#INTC_otc"
    BA_otc = "#BA_otc"
    JNJ_otc = "#JNJ_otc"
    PFE_otc = "#PFE_otc"
    XOM_otc = "#XOM_otc"
    AXP_otc = "#AXP_otc"
    MCD_otc = "#MCD_otc"
    CSCO_otc = "#CSCO_otc"
    VISA_otc = "#VISA_otc"
    CITI_otc = "#CITI_otc"
    FDX_otc = "#FDX_otc"
    TWITTER_otc = "#TWITTER_otc"
    BABA_otc = "#BABA_otc"

    def is_otc(self) -> bool:
        return self.endswith("_otc")

    def __new__(cls, value: str) -> "Asset":
        for member in cls:
            if member.value == value:
                return member

        obj = str.__new__(cls, value)
        obj._name_ = value
        obj._value_ = value
        return obj

    @classmethod
    def _missing_(cls, value: str) -> "Asset":  # type: ignore
        """
        Create a new Asset instance for unknown values dynamically.
        Does NOT modify class attributes (avoids AttributeError).
        """
        obj = str.__new__(cls, value)
        obj._name_ = value
        obj._value_ = value

        # Register in value map so that repeated calls return the same object
        cls._value2member_map_[value] = obj

        return obj

    @classmethod
    def __get_validators__(cls):  # noqa: ANN206
        yield cls.validate

    @classmethod
    def validate(cls, value: typing.Any) -> "Asset":
        if isinstance(value, cls):
            return value
        if isinstance(value, str):
            return cls(value)
        raise TypeError(f"Invalid type for Asset: {type(value)}")

    @classmethod
    def __get_pydantic_core_schema__(cls, _source_type, _handler):  # noqa: ANN001, ANN206
        from pydantic_core import core_schema  # noqa: F811, PLC0415, RUF100

        return core_schema.no_info_after_validator_function(cls, core_schema.str_schema())


class AuthorizationData(pydantic.BaseModel):
    session: str
    is_demo: typing.Annotated[IsDemo, pydantic.Field(..., alias="isDemo")]
    uid: int
    platform: int
    is_fast_history: typing.Annotated[bool, pydantic.Field(..., alias="isFastHistory")]
    is_optimized: typing.Annotated[bool, pydantic.Field(..., alias="isOptimized")]


class SuccessUpdateBalance(pydantic.BaseModel):
    is_demo: typing.Annotated[IsDemo, pydantic.Field(..., alias="isDemo")]
    balance: float


class UpdateHistoryFastEvent(pydantic.BaseModel):
    asset: Asset
    period: int
    history: list[list[float]]


class UpdateStreamItem(pydantic.BaseModel):
    asset: Asset
    timestamp: float
    value: float


UpdateStreamTypeAdapter = pydantic.TypeAdapter(list[UpdateStreamItem])


class OpenPendingOrderRequestOpenType(enum.IntEnum):
    TIME = 0
    PRICE = 1


class Command(enum.IntEnum):
    PUT = 0
    CALL = 1


class Deal(pydantic.BaseModel):
    id: uuid.UUID
    open_time: typing.Annotated[datetime.datetime, pydantic.Field(..., alias="openTime")]
    close_time: typing.Annotated[datetime.datetime, pydantic.Field(..., alias="closeTime")]
    open_timestamp: typing.Annotated[float, pydantic.Field(..., alias="openTimestamp")]
    close_timestamp: typing.Annotated[float | None, pydantic.Field(None, alias="closeTimestamp")]
    refund_time: typing.Annotated[datetime.datetime | None, pydantic.Field(None, alias="refundTime")]
    refund_timestamp: typing.Annotated[int | None, pydantic.Field(None, alias="refundTimestamp")]
    uid: int
    amount: float
    profit: float
    percent_profit: typing.Annotated[float, pydantic.Field(..., alias="percentProfit")]
    percent_loss: typing.Annotated[float, pydantic.Field(..., alias="percentLoss")]
    open_price: typing.Annotated[float, pydantic.Field(..., alias="openPrice")]
    close_price: typing.Annotated[float | None, pydantic.Field(None, alias="closePrice")]
    command: Command
    asset: Asset
    is_demo: typing.Annotated[IsDemo, pydantic.Field(..., alias="isDemo")]
    copy_ticket: typing.Annotated[str, pydantic.Field(..., alias="copyTicket")]
    open_ms: typing.Annotated[int | None, pydantic.Field(None, alias="openMs")]
    close_ms: typing.Annotated[int | None, pydantic.Field(None, alias="closeMs")]
    option_type: typing.Annotated[int | None, pydantic.Field(None, alias="optionType")]
    is_rollover: typing.Annotated[bool | None, pydantic.Field(None, alias="isRollover")]
    is_copy_signal: typing.Annotated[bool, pydantic.Field(..., alias="isCopySignal")]
    is_ai: typing.Annotated[bool | None, pydantic.Field(None, alias="isAI")]
    currency: str
    amount_usd: typing.Annotated[float | None, pydantic.Field(None, alias="amountUSD")]
    request_id: typing.Annotated[int | None, pydantic.Field(None, alias="requestId")]


DealListTypeAdapter = pydantic.TypeAdapter(list[Deal])


class OrderAction(enum.StrEnum):
    CALL = "call"
    PUT = "put"


class OpenOrderRequest(pydantic.BaseModel):
    model_config = pydantic.ConfigDict(validate_by_name=True, validate_by_alias=True)

    asset: Asset
    amount: int
    action: OrderAction
    is_demo: typing.Annotated[IsDemo, pydantic.Field(..., alias="isDemo")]
    request_id: typing.Annotated[int, pydantic.Field(..., alias="requestId")]
    option_type: typing.Annotated[int, pydantic.Field(..., alias="optionType")]
    time: int


class CopySignalRequest(pydantic.BaseModel):
    symbol: Asset
    amount: int
    expired_at: typing.Annotated[int, pydantic.Field(..., alias="expiredAt")]
    action: OrderAction
    is_demo: typing.Annotated[IsDemo, pydantic.Field(..., alias="isDemo")]
    request_id: typing.Annotated[int, pydantic.Field(..., alias="requestId")]
    created_at: typing.Annotated[int, pydantic.Field(..., alias="createdAt")]
    timeframe: int
    signal_id: typing.Annotated[str, pydantic.Field(..., alias="signalId")]


class OpenPendingOrderRequest(pydantic.BaseModel):
    open_type: typing.Annotated[OpenPendingOrderRequestOpenType, pydantic.Field(..., alias="openType")]
    amount: int
    asset: Asset
    open_time: typing.Annotated[str, pydantic.Field(..., alias="openTime")]
    open_price: typing.Annotated[int, pydantic.Field(..., alias="openPrice")]
    timeframe: int
    min_payout: typing.Annotated[int, pydantic.Field(..., alias="minPayout")]
    command: Command


class ChangeSymbolRequest(pydantic.BaseModel):
    asset: Asset
    period: int


class SuccessCloseOrder(pydantic.BaseModel):
    profit: float
    deals: list[Deal]
